from ..classes.pareto import pareto
import random

def Pareto(model, objectives, objdirec, runs, GetPoints=True, tol=1e-10):
    """ pre: objectives = [["reac"],{"reac2":x}]
       post: turning points of Pareto front """
    state = model.GetState()
    rv = pareto()
    model.SetObjDirec(objdirec)
    anchor = []
    for obj in objectives:
        model.ZeroObjective()
        model.SetObjective(obj)
        model.Solve(PrintStatus=False)
        anchor.append(model.GetObjVal())
    print(anchor)   
    if len(anchor) == len(objectives):
        for n in range(runs):
            model.ZeroObjective()
            coef = []
            for b in range(len(objectives)):
                coef.append(random.random())
            sumcoef = sum(coef)
            for b in range(len(objectives)):
                try: 
                    coef[b] = coef[b]/anchor[b]/sumcoef
                except ZeroDivisionError:
                    print("Zero Division error at %s" % objectives[b])
                    continue
            objdic = {}
            for b in range(len(objectives)):
                thisobjdic = {}
                if isinstance(objectives[b],list):
                    for reac in objectives[b]:
                        thisobjdic[reac] = coef[b]
                elif isinstance(objectives[b],dict):
                    for reac in objectives[b]:
                        thisobjdic[reac] = objectives[b][reac]*coef[b]
                for r in thisobjdic:
                    if r in objdic.keys():
                        objdic[r] += thisobjdic[r]
                    else:
                        objdic[r] = thisobjdic[r]
            print(objdic)
            model.SetObjective(objdic)
            model.Solve(PrintStatus=False)
            sol = model.GetSol(IncZeroes=True)
            for b in range(len(objectives)):
                sol["coef"+str(b+1)] = coef[b]
                if isinstance(objectives[b],list):
                    sol["Obj"+str(b+1)] = sol[objectives[b][0]]
                elif isinstance(objectives[b],dict):
                    objsol = 0
                    for reac in objectives[b]:
                        objsol += sol[reac]*objectives[b][reac]
                    sol["Obj"+str(b+1)] = objsol
            print(sol)
            rv = pareto(rv.UpdateFromDic(sol))
    model.SetState(state)
    if len(anchor) == len(objectives) and GetPoints:
        return rv.GetParetoPoints(tol)
    else:
        return rv





def LinearMOMA(model, refflux):
    model.MinDiffFromFlux(refflux, reacs=refflux.keys(), it=0, cleanup=True)

def MinDiffFromFlux(model, fluxdist, reacs=None, it=0, cleanup=True):
    """ pre: fluxdist = {reac:flux_val} """
    state = model.GetState()
    obj = {}
    if not reacs:
        reacs = fluxdist.keys()
    for r in reacs:
        rowname = r+"_"+str(it)+"_mindifffromflux"
        colplus = r+"_"+str(it)+"_mindifffromflux+"
        colminus = r+"_"+str(it)+"_mindifffromflux-"
        model.AddMetabolite(rowname)
        model.GetReaction(r).add_metabolites({model.GetMetabolite(rowname):1})
        model.AddReaction(colplus,{model.GetMetabolite(rowname):1})
        model.AddReaction(colminus,{model.GetMetabolite(rowname):-1})
        model.SetMetBounds(rowname, fluxdist[r], fluxdist[r])
        obj[colplus] = 1
        obj[colminus] = 1
    model.ZeroObjective()
    model.SetObjective(obj)
    model.SetObjDirec("Min")
    model.Solve(False)
    ### cleanup
    if cleanup:
        model.CleanUpTempVar("_mindifffromflux")
    model.SetState(state,IncSol=False)

def CleanUpTempVar(model, var):
    for row in model.Metabolites():
        if var in row:
            model.SetMetBounds(row, 0, 0)
            #model.DelMetabolite(row)
    for col in model.Reactions():
        if var in col:
            model.DelReaction(col)
    if len(model.solution.x_dict) != 0:
        for x in list(model.solution.x_dict.keys()):
            if var in x:
                del model.solution.x_dict[x]
    #if len(model.solution.x) != 0:
    #    model.solution.x = model.solution.x[:len(model.reactions)]
    if len(model.solution.y_dict) != 0:
        for y in list(model.solution.y_dict.keys()):
            if var in y:
                del model.solution.y_dict[y]
    #if len(model.solution.y) != 0:
    #    model.solution.y = model.solution.y[:len(model.metabolites)]
    ### x and y attributes of model.solution does not have a 
    ### setter property; thus, their values cannot be changed. 
    ### To access a similar data, use x_dict and y_dict. 

def MOMA(model, refflux):
    state = model.GetState()
    model.ZeroObjective()
    model.SetObjDirec("Min")
    for reac in refflux:
        model.SetQuadraticObjective({reac:1})
        model.SetObjective({reac:-refflux[reac]})
    model.Solve(False)
    model.SetState(state,IncSol=False)





import builtins as exceptions
#many unsupported attribute of types lib in python 3
import types
import re, math 
from collections import defaultdict
import numpy
import scipy
#import metabolite

import cobra
from cobra import Metabolite, Reaction, Gene
from cobra.flux_analysis import deletion, moma, phenotype_phase_plane
from cobra.core.solution import get_solution
#from cobra.manipulation import modify

#from ..analysis import FCA, Pareto, RWFM, MOMA, ROOM, GeometricFBA, MinSolve
#from ..analysis import Graph, FluxSum, FVA, MinSolve, Scan
#from ..manipulation import Reversible
#from ..classes.flux import flux
#rom ..io import Network

class Reaction(cobra.Reaction):
	def __init__(self,id=None, name='', subsystem='', lower_bound=0.,upper_bound=1000., objective_coefficient=0.):
		super().__init__(id=id,name=name,subsystem=subsystem,lower_bound=lower_bound,upper_bound=upper_bound)
		#self.objective_coefficient = objective_coefficient
		return
	def __str__(self):
		return


import re
import pandas
import cobra
from fractions import Fraction


def ReadExcel(excel_file, parse="cobra_string", Print=False):
    """ parse = "cobra_string" | "cobra_position"
    cobra_string
    % INPUT
    % fileName      xls spreadsheet, with one 'Reaction List' and one 'Metabolite List' tab
    %
    % 'Reaction List' tab: Required headers (case sensitive):
    %   'Abbreviation'      HEX1
    %   'Description'       Hexokinase
    %   'Reaction'          1 atp[c] + 1 glc-D[c] --> 1 adp[c] + 1 g6p[c] + 1 h[c]
    %   'GPR'               (3098.3) or (80201.1) or (2645.3) or ...
    %   'Genes'             2645.1,2645.2,2645.3,...  (optional)
    %   'Proteins'          Flj22761.1, Hk1.3, Gck.2,...  (optional)
    %   'Subsystem'         Glycolysis
    %   'Reversible'        0 (false) or 1 (true)
    %   'Lower bound'       0
    %   'Upper bound'       1000
    %   'Objective'         0   (optional)
    %   'Confidence Score'  0,1,2,3,4
    %   'EC Number'         2.7.1.1,2.7.1.2
    %   'Notes'             'Reaction also associated with EC 2.7.1.2' (optional)
    %   'References'        PMID:2043117,PMID:7150652,...   (optional)
    %
    % 'Metabolite List' tab: Required headers (case sensitive): (needs to be complete list of metabolites, i.e., if a metabolite appears in multiple compartments it has to be represented in multiple rows. Abbreviations need to overlap with use in Reaction List
    %   'Abbreviation'      glc-D or glc-D[c]
    %   'Description'       D-glucose
    %   'Neutral formula'   C6H12O6
    %   'Charged formula'   C6H12O6
    %   'Charge'            0
    %   'Compartment'       cytosol
    %   'KEGG ID'           C00031
    %   'PubChem ID'        5793
    %   'ChEBI ID'          4167
    %   'InChI string'      InChI=1/C6H12O6/c7-1-2-3(8)4(9)5(10)6(11)12-2/h2-11H,1H2/t2-,3-,4+,5-,6?/m1/s1
    %   'SMILES'            OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O
    %   'HMDB ID'           HMDB00122
    %
    % OPTIONAL INPUT (may be required for input on unix macines)
    % biomassRxnEquation        .xls may have a 255 character limit on each cell,
    %                           so pass the biomass reaction separately if it hits this maximum.
    %
    % OUTPUT
    % model         COBRA Toolbox model

    cobra_position
    % INPUT
    % excel_file      xls spreadsheet, with one 'reactions' and one 'metabolites' tab
    %
    % 'reactions' tab: Required headers:
    %               col 0     Abbreviation     HEX1
    %               col 1     Description      Hexokinase
    %               col 2     Reaction         1 atp[c] + 1 glc-D[c] --> 1 adp[c] + 1 g6p[c] + 1 h[c]
    %               col 3     GPR              b0001
    %               col 4     Genes            b0001 (optional: column can be empty)
    %               col 5     Proteins         AlaS (optional: column can be empty)
    %               col 6     Subsystem        Glycolysis
    %               col 7     Reversible       0
    %               col 8     Lower bound      0
    %               col 9     Upper bound      1000
    %               col 10    Objective        0    (optional: column can be empty)
    %               col 11    Confidence Score 0,1,2,3,4
    %               col 12    EC. Number       1.1.1.1
    %               col 13    Notes            N/A  (optional: column can be empty)
    %               col 14    References       PMID: 1111111  (optional: column can be empty)
    %
    % 'metabolites' tab: Required headers: needs to be complete list of metabolites, i.e., if a metabolite appears in multiple compartments it has to be represented in multiple rows. Abbreviations needs to overlap with use in Reaction List
    %               col 0     Abbreviation
    %               col 1     Description
    %               col 2     Neutral formula
    %               col 3     Charge formula
    %               col 4     Charge
    %               col 5     Compartment
    %               col 6     KEGG ID
    %               col 7     PubChem ID
    %               col 8     ChEBI ID
    %               col 9     InChI string
    %               col 10    SMILES
    %               col 11    HMDB ID
    %
    %
    % OUTPUT
    % model         cobrapy model """

    excel = pandas.ExcelFile(excel_file)
    for sheet in excel.sheet_names:
        if sheet == "Reaction List" and parse == "cobra_string":
            reactions = excel.parse(sheet,index_col=None)
        elif 'reaction' in sheet.lower():
            reactions = excel.parse(sheet,index_col=None)
        if sheet == "Metabolite List" and parse == "cobra_string":
            metabolites = excel.parse(sheet,index_col=None)
        elif 'metabolite' in sheet.lower():
            metabolites = excel.parse(sheet,index_col=None)

    cobra_reaction_position = ['Abbreviation','Description','Reaction','GPR','Genes','Proteins','Subsystem','Reversible','Lower bound','Upper bound','Objective','Confidence Score','EC Number','Notes','References']
    cobra_metabolite_position = ['Abbreviation','Description','Neutral formula','Charged formula','Charge','Compartment','KEGG ID','PubChem ID','ChEBI ID','InChI string','SMILES','HMDB ID']

    if parse == "cobra_position":
        if len(reactions.columns) > 15:
            reactions = reactions.iloc[:,:15]
            reactions.columns = cobra_reaction_position
        else:
            reactions.columns = cobra_reaction_position[:len(reactions.columns)]
        if len(metabolites.columns) > 12:
            metabolites = metabolites.iloc[:,:12]
            metabolites.columns = cobra_metabolite_position
        else:
            metabolites.columns = cobra_metabolite_position[:len(metabolites.columns)]

    model = cobra.Model()
    metabolite_dic = {}
    element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")
    for met in metabolites.index:
        met_row = metabolites.loc[met]  # pandas.Series of the metabolites
        met_id = str(met_row['Abbreviation'])
        #met_name = str(met_row[1]) if pandas.notnull(met_row[1]) else None
        met_name = str(met_row['Description']) if ('Description' in met_row.index) and pandas.notnull(met_row['Description']) else None

        if ('Charged formula' in met_row.index) and pandas.notnull(met_row['Charged formula']):
            met_formula = str(met_row['Charged formula'])
        elif ('Neutral formula' in met_row.index) and pandas.notnull(met_row['Neutral formula']):
            if ('Charge' in met_row.index) and pandas.notnull(met_row['Charge']):
                met_formula = ''
                tmp_formula = str(met_row['Neutral formula'])
                tmp_formula = tmp_formula.replace("*", "")
                parsed = element_re.findall(tmp_formula)
                for (element, count) in parsed:
                    if element != "H":
                        met_formula += element + str(count)
                    else:
                        if count == '':
                            count = 1
                        count = float(count)
                        if count.is_integer():
                            count = int(count)
                        charge = float(met_row['Charge'])
                        if charge.is_integer():
                            charge = int(charge)
                        count += charge
                        if count == 1:
                            met_formula += element
                        elif count != 0:
                            met_formula += element + str(count)
            else:
                met_formula = None
        else:
            met_formula = None
        met_compartment = str(met_row['Compartment']) if 'Compartment' in met_row.index and pandas.notnull(met_row['Compartment']) else None
        metabolite = cobra.Metabolite(met_id, formula=met_formula, name=met_name, compartment=met_compartment)
        if ('Charge' in met_row.index) and pandas.notnull(met_row['Charge']):
            metabolite.charge = float(met_row['Charge'])
            if metabolite.charge.is_integer():
                metabolite.charge = int(metabolite.charge)
        if ('KEGG ID' in met_row.index) and pandas.notnull(met_row['KEGG ID']): metabolite.kegg_id = met_row['KEGG ID']
        if 'PubChem ID' in met_row.index and pandas.notnull(met_row['PubChem ID']): metabolite.pubchem_id = str(met_row['PubChem ID'])
        if 'ChEBI ID' in met_row.index and pandas.notnull(met_row['ChEBI ID']): metabolite.chebi_id = str(met_row['ChEBI ID'])
        if 'InChI string' in met_row.index and pandas.notnull(met_row['InChI string']): metabolite.inchi_id = str(met_row['InChI string'])
        if 'SMILES' in met_row.index and pandas.notnull(met_row['SMILES']): metabolite.smiles = str(met_row['SMILES'])
        if 'HMDB ID' in met_row.index and pandas.notnull(met_row['HMDB ID']): metabolite.hmdb_id = str(met_row['HMDB ID'])
        metabolite_dic[met_id] = metabolite
        if Print:
            print(metabolite.id)

    for reac in reactions.index:
        reac_row = reactions.loc[reac]
        reaction = cobra.Reaction(str(reac_row['Abbreviation']))

        stoichiometry_list = []
        metabolite_list = []
        new_metabolite = True
        product = False
        compartment = ''
        reaction_string = reac_row['Reaction'].split(' ')
        for s in reaction_string:
            if s.startswith('['):
                if s.endswith(']:'):
                    compartment = s[:-1]
                else:
                    compartment = s
            elif s == ":" or s == "":
                pass
            elif s == "+":
                new_metabolite = True
            elif s == "->" or s == "-->" or s == "=>" or s == u'\u2192':
                direction = 'LEFT-TO-RIGHT'
                product = True
                new_metabolite = True
            elif s == "<>" or s =="<=>" or s == "<==>" or s == "<->" or s == u'\u2194':
                direction = 'REVERSIBLE'
                product = True
                new_metabolite = True
            elif s == "<-" or s == "<--" or s == "<=" or s == u'\u2190':
                direction = 'RIGHT-TO-LEFT'
                product = True
                new_metabolite = True
            else:
                try:
                    if s[0] == "(" and s[-1] == ")":
                        s = s[1:-1]
                    s = Fraction(s)
                    s = float(s)
                    if not product:
                        s = -s
                    stoichiometry_list.append(s)
                    new_metabolite = False
                except ValueError:
                    metabolite_list.append(s + compartment)
                    if new_metabolite:
                        if not product:
                            stoichiometry_list.append(-1)
                        else:
                            stoichiometry_list.append(1)
                        new_metabolite = True
        stoi_dic = {}
        for b in range(len(metabolite_list)):
            met_id = str(metabolite_list[b])
            try:
                met = metabolite_dic[met_id]
            except KeyError:    # try case-insensitive
                met_key = []
                for key in metabolite_dic:
                    if key.lower() == met_id.lower():
                        met_key.append(key)
                if len(met_key) == 1:
                    met = metabolite_dic[met_key[0]]
                else:
                    print(met_id + " added to metabolite list in " + str(reac_row['Abbreviation']) + " in row " + str(reac+2))
                    additional_metabolite = cobra.Metabolite(met_id)
                    met = metabolite_dic[met_id] = additional_metabolite
            stoi_dic[met] = stoichiometry_list[b]
        reaction.add_metabolites(stoi_dic)
        #
        reaction.lower_bound = -1000
        reaction.upper_bound = 1000
        if direction == 'LEFT-TO-RIGHT':
            reaction.lower_bound = 0.0
        elif direction == 'RIGHT-TO-LEFT':
            reaction.upper_bound = 0.0

        if ('Description' in reac_row.index) and pandas.notnull(reac_row['Description']): reaction.name = str(reac_row['Description'])
        if ('GPR' in reac_row.index) and pandas.notnull(reac_row['GPR']): reaction.gene_reaction_rule = str(reac_row['GPR'])
        if ('Proteins' in reac_row.index) and pandas.notnull(reac_row['Proteins']): reaction.proteins = str(reac_row['Proteins'])
        if ('Subsystem' in reac_row.index) and pandas.notnull(reac_row['Subsystem']): reaction.subsystem = str(reac_row['Subsystem'])
        if ('Lower bound' in reac_row.index) and pandas.notnull(reac_row['Lower bound']): reaction.lower_bound = float(reac_row['Lower bound'])
        if ('Upper bound' in reac_row.index) and pandas.notnull(reac_row['Upper bound']): reaction.upper_bound = float(reac_row['Upper bound'])
        if ('Confidence Score' in reac_row.index) and pandas.notnull(reac_row['Confidence Score']): reaction.confidence_score = int(reac_row['Confidence Score'])
        if ('EC Number' in reac_row.index) and pandas.notnull(reac_row['EC Number']): reaction.ec_number = str(reac_row['EC Number'])
        if ('Notes' in reac_row.index) and pandas.notnull(reac_row['Notes']): reaction.notes = {"notes":str(reac_row['Notes'])}
        if ('References' in reac_row.index) and pandas.notnull(reac_row['References']): reactions.references = str(reac_row['References'])

        model.add_reaction(reaction)
        if ('Objective' in reac_row.index) and pandas.notnull(reac_row['Objective']): reaction.objective_coefficient = float(reac_row['Objective'])
        if Print:
            print(reaction.reaction)
 
    return model

def WriteExcel(model,filename,excel_format="cobra"):
    """ excel_format = "cobra" | "cobra_old" """
    r_dict = {'Abbreviation':{},'Description':{},'Reaction':{},'GPR':{},'Genes':{},'Proteins':{},'Subsystem':{},'Reversible':{},'Lower bound':{},'Upper bound':{},'Objective':{},'Confidence Score':{},'EC Number':{},'Notes':{},'References':{}}
    for r in model.reactions:
        r_dict['Abbreviation'][r.id] = r.id
        r_dict['Description'][r.id] = getattr(r,'name','')
        r_dict['Reaction'][r.id] = r.reaction.replace('<--','<=>')
        r_dict['GPR'][r.id] = getattr(r,'gene_reaction_rule','')
        genes = ''
        if hasattr(r,'genes'):
            for gene in r.genes:
                genes = genes + gene.id + ' '
            genes = genes[:-1]
        r_dict['Genes'][r.id] = genes
        r_dict['Proteins'][r.id] = getattr(r,'proteins','')
        r_dict['Subsystem'][r.id] = getattr(r,'subsystem','')
        r_dict['Reversible'][r.id] = 1 if r.reversibility else 0
        r_dict['Lower bound'][r.id] = r.lower_bound
        r_dict['Upper bound'][r.id] = r.upper_bound
        r_dict['Objective'][r.id] = getattr(r,'objective_coefficient','')
        r_dict['Confidence Score'][r.id] = getattr(r,'confidence_score','')
        r_dict['EC Number'][r.id] = getattr(r,'ec_number','')
        r_dict['Notes'][r.id] = str(getattr(r,'notes','')) if str(getattr(r,'notes','')) != '{}' else ''
        r_dict['References'][r.id] = getattr(r,'references','')

    m_dict = {'Abbreviation':{},'Description':{},'Neutral formula':{},'Charged formula':{},'Charge':{},'Compartment':{},'KEGG ID':{},'PubChem ID':{},'ChEBI ID':{},'InChI string':{},'SMILES':{},'HMDB ID':{}}
    element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")
    for m in model.metabolites:
        m_dict['Abbreviation'][m.id] = m.id
        m_dict['Description'][m.id] = getattr(m,'name','')
        neutral_formula = ''
        if m.charge != None and m.formula != None:
            tmp_formula = str(m.formula).replace('*', '')
            parsed = element_re.findall(tmp_formula)
            for (element, count) in parsed:
                if element != 'H':
                    neutral_formula += element + str(count)
                else:
                    if count == '':
                        count = 1
                    count = float(count)
                    if count.is_integer():
                        count = int(count)
                    count -= m.charge
                    if count == 1:
                        neutral_formula += element
                    elif count != 0:
                        neutral_formula += element + str(count)
        m_dict['Neutral formula'][m.id] = neutral_formula
        m_dict['Charged formula'][m.id] = str(getattr(m,'formula','')) if getattr(m,'formula','') != None else ''
        m_dict['Charge'][m.id] = getattr(m,'charge','')
        m_dict['Compartment'][m.id] = getattr(m,'compartment','') if getattr(m,'compartment','') != None else ''
        m_dict['KEGG ID'][m.id] = getattr(m,'kegg_id','')
        m_dict['PubChem ID'][m.id] = getattr(m,'pubchem_id','')
        m_dict['ChEBI ID'][m.id] = getattr(m,'chebi_id','')
        m_dict['InChI string'][m.id] = getattr(m,'inchi_id','')
        m_dict['SMILES'][m.id] = getattr(m,'smiles','')
        m_dict['HMDB ID'][m.id] = getattr(m,'hmdb_id','')

    reactions = pandas.DataFrame(r_dict, columns=['Abbreviation','Description','Reaction','GPR','Genes','Proteins','Subsystem','Reversible','Lower bound','Upper bound','Objective','Confidence Score','EC Number','Notes','References'])
#    reactions = reactions.loc[:, ~reactions.isin(['',None]).all()]

    metabolites = pandas.DataFrame(m_dict, columns=['Abbreviation','Description','Neutral formula','Charged formula','Charge','Compartment','KEGG ID','PubChem ID','ChEBI ID','InChI string','SMILES','HMDB ID'])
#    metabolites = metabolites.loc[:, ~metabolites.isin(['',None]).all()]

    excel_file = pandas.ExcelWriter(filename)
    if excel_format == "cobra_old":
        reactions.to_excel(excel_file,'reactions',index=False)
    elif excel_format == "cobra":
        reactions.to_excel(excel_file,'Reaction List',index=False)
    if excel_format == "cobra_old":
        metabolites = metabolites.drop('HMDB ID', 1)
        metabolites.to_excel(excel_file,'metabolites',index=False)
    elif excel_format == "cobra":
        metabolites.to_excel(excel_file,'Metabolite List',index=False)
    excel_file.save()





def FluxSum(model,met,tol=1e-10):
    sol = model.GetSol(IncZeroes=True,AsID=False)
    if model.GetSol(IncZeroes=False) == {}:
        state = model.GetState()
        model.Solve(PrintStatus=False)
        sol = model.GetSol(IncZeroes=True,AsID=False)
        model.SetState(state)
    prodreacs = model.ProducedBy(met)
    consreacs = model.ConsumedBy(met)
    prod = 0.0
    for reac in prodreacs.keys():
        reacprod = sol[reac]*prodreacs[reac]
        if reacprod > 0:
            prod += reacprod
    cons = 0.0
    for reac in consreacs.keys():
        reaccons = sol[reac]*consreacs[reac]
        if reaccons < 0:
            cons += reaccons
    if abs(prod + cons) < tol:
        rv = (abs(prod)+abs(cons))/2.0
        return rv
    else:
        print (met + " is external / not balanced")
        print ("net balance = ", prod+cons)


def ProducedBy(model,met,FixBack=True):
    met = model.GetMetabolite(met)
    reactions = model.InvolvedWith(met)
    rv = {}
    for reac in reactions:
        stoi = model.InvolvedWith(reac)[met]
        if reac.reversibility or (stoi > 0):
            rv[reac] = stoi
    return rv

def ConsumedBy(model,met,FixBack=True):
    met = model.GetMetabolite(met)
    reactions = model.InvolvedWith(met)
    rv = {}
    for reac in reactions:
        stoi = model.InvolvedWith(reac)[met]
        if reac.reversibility or (stoi < 0):
            rv[reac] = stoi
    return rv



import builtins as exceptions
#many unsupported attribute of types lib in python 3
import types
import re, math 
from collections import defaultdict
import numpy
import scipy

import cobra
from cobra import Metabolite, Reaction, Gene
from cobra.flux_analysis import deletion, moma, phenotype_phase_plane
from cobra.core.solution import get_solution
#from cobra.manipulation import modify

#from ..analysis import FCA, Pareto, RWFM, MOMA, ROOM, GeometricFBA, MinSolve
#from ..analysis import Graph, FluxSum, FVA, MinSolve, Scan
#from ..manipulation import Reversible
#from ..classes.flux import flux
#from ..io import Network

class Metabolite(cobra.Metabolite):
	"""Metabolite is a class inherited from the cobra Metabolite class that holds information
	that can be used in the scobra reaction class.  It differs forom the original cobra
	Metabolite class with 3 new states:

	Parameters
	----------
	inchi :
	smiles :
	molecular_weights : float
		the combined weight of the molecules in the metabolite
	"""

	def __init__(self,id=None, formula=None, name="",charge=None, compartment=None,inchi=None,smiles=None,molecular_weights=None,comments=""):
		super().__init__(id=id, formula=formula, name=name,charge=charge, compartment=compartment)
		self.inchi = inchi
		self.smiles = smiles
		self.molecular_weights = molecular_weights
		self.comments = comments


import pandas
import math
try: 
    
    import matplotlib
    matplotlib.use("TkAgg")
    import matplotlib.pyplot
    
except ImportError: 
    pass
from scipy import stats



class matrix(pandas.DataFrame):

    def __init__(self, *args, **kwargs):
        super(matrix, self).__init__(*args, **kwargs)

    def Copy(self):
        return matrix(self.copy())

    @classmethod
    def FromFile(self, path, file_format=None, **kwargs):
        """ file_format = "csv" | "excel" | "pickle" """
        if file_format == "csv" or path.endswith('.csv') or path.endswith('.txt'):
            return matrix(pandas.read_csv(path, **kwargs))
        elif file_format == "excel" or file_format == "xls" or file_format == "xlsx" or path.endswith('.xls') or path.endswith('.xlsx'):
            return matrix(pandas.read_excel(path, **kwargs))
        elif file_format == "pickle" or path.endswith('.p') or path.endswith('.pkl') or path.endswith('.pickle'):
            return matrix(pandas.read_pickle(path))

    def ToFile(self, path, file_format=None, **kwargs):
        """ file_format = "csv" | "excel" | "pickle" """
        if file_format == "csv" or path.endswith('.csv') or path.endswith('.txt'):
            self.to_csv(path, **kwargs)
        elif file_format == "excel" or file_format == "xls" or file_format == "xlsx" or path.endswith('.xls') or path.endswith('.xlsx'):
            self.to_excel(path, **kwargs)
        elif file_format == "pickle" or path.endswith('.p') or path.endswith('.pkl') or path.endswith('.pickle'):
            self.to_pickle(path)

    def Plot(self,x,y=None,**kwargs):
        if y == None:
            self[x].plot(**kwargs)
        else:
            self.plot(kind="scatter",x=x,y=y)
        pyplot.show()

    def VaryReacs(self,tol=1e-10):
        varymtx = self.Copy()
        fixreacs = []
        for reac in self.columns:
            vary = False
            col = list(self[reac])
            for val in col:
                if abs(val-col[0]) > tol:
                    vary = True
            if vary == False:
                fixreacs.append(reac)
        for reac in fixreacs:
            varymtx=varymtx.drop(reac,1)
        return matrix(varymtx)

    def FixedReacs(self,tol=1e-10):
        fixedmtx = self.Copy()
        varyreacs = []
        for reac in self.columns:
            vary = False
            col = list(self[reac])
            for val in col:
                if abs(val-col[0]) > tol:
                    vary = True
            if vary == True:
                varyreacs.append(reac)
        for reac in varyreacs:
            fixedmtx=fixedmtx.drop(reac,1)
        return matrix(fixedmtx)

    def NonZeroes(self,tol=1e-10):
        rv = self.Copy()
        zeroreacs = []
        for reac in self.columns:
            zero = True
            col = list(self[reac])
            for val in col:
                if abs(val) > tol:
                    zero = False
            if zero:
                zeroreacs.append(reac)
        for reac in zeroreacs:
            rv = rv.drop(reac,1)
        return matrix(rv)

    def ZeroReac(self,tol=1e-10):
        rv = []
        for reac in self.columns:
            zero = True
            col = list(self[reac])
            for val in col:
                if abs(val) > tol:
                    zero = False
            if zero:
                rv.append(reac)
        return rv

    def AverageFlux(self,IncZeroes=False,AsMtx=False,tol=1e-10):
        rv = {}
        for reac in self.columns:
            rv[reac] = float(sum(self[reac]))/len(self.index)
        if not IncZeroes:
            for reac in list(rv.keys()):
                if abs(rv[reac]) < tol:
                    del rv[reac]
        if AsMtx:
           rv = matrix({"lp_sol":rv})
        return rv

    def FluxRange(self,IncZeroes=False,AsMtx=False,tol=1e-10):
        from .fva import fva
        rv = fva(bounds=self.bounds)
        for reac in self.columns:
            rv[reac] = (min(self[reac]),max(self[reac]))
        if not IncZeroes:
            rv = rv.Allowed(tol=tol)
        if AsMtx:
           rv = rv.AsMtx()
        return rv

    def DicUpdate(self,dic,row=None):
        s = pandas.Series(dic,name=row)
        rv = self.append(s,ignore_index=row==None)
        rv = rv.fillna(0)
        return matrix(rv)

    def UpdateFromDic(self,dic,row=None):
        return self.DicUpdate(dic=dic,row=row)

#    def Tree(self,incobjval=False,tol=1e-10):
#        temp = self.VaryReacs(tol=tol)
#        if incobjval == False and 'ObjVal' in temp.columns:
#            temp = temp.drop('ObjVal',1)
#        temp = temp.T
#        simmtx = temp.RowDiffMtx(lambda x,y: 1-abs(Stats.Pearson_r(x,y)))
#        t = simmtx.ToNJTree()
#        return t

    def ResponseCoef(self,scanvar=None,incobjval=False,tol=1e-10):
        temp = self.VaryReacs(tol=tol)
        if scanvar == None:
            var = temp.iloc[:,0]
        else:
            var = temp[scanvar]
        if incobjval == False and 'ObjVal' in temp.columns:
            temp = temp.drop('ObjVal')
        rv = {}
        for reac in temp.columns:
            rv[reac] = 1-abs(stats.pearsonr(var,temp[reac])[0])
        return rv

    def PrintResponseCoef(self,lo=0,hi=1,rc=None,f=None,Sort="value",sortabs=True,reverse=True,scanvar=None,incobjval=False,tol=1e-10):
        if rc == None:
            rc = self.ResponseCoef(scanvar=scanvar,incobjval=incobjval,tol=tol)
        self.PrintSorted(rc, lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs, reverse=reverse)

    def FluxCorrCoefMtx(self,absolute=False,tol=1e-10):
        vary = self.VaryReacs(tol=tol)
        lenvary = len(vary.columns)
        rv = matrix()
        rv.columns = list(vary.columns)
        for reac in vary.columns:
            rv.loc[reac,:] = 0
        for r1 in range(lenvary):
            for r2 in range(r1, lenvary):
                corrcoef = stats.pearsonr(vary[r1],vary[r2])[0]
                if absolute:
                    corrcoef = abs(corrcoef)
                rv.iloc[r1,r2] = rv.iloc[r2,r1] = corrcoef
        return rv


    def FluxCorrCoef(self,reac1,reac2=None,absolute=False,tol=1e-10):
        temp = self.VaryReacs(tol=tol)
        if reac2 != None:
            if absolute:
                return abs(stats.pearsonr(temp[reac1],temp[reac2]))[0]
            else:
                return stats.pearsonr(temp[reac1],temp[reac2])[0]
        else:
            rv = {}
            for reac in temp.columns:
                if absolute:
                    rv[reac] = abs(stats.pearsonr(temp[reac1],temp[reac]))[0]
                else:
                    rv[reac] = stats.pearsonr(temp[reac1],temp[reac])[0]
            return rv

    def PrintFluxCorrCoef(self,reac,lo=0,hi=1,f=None,Sort="value",sortabs=True,reverse=True,absolute=False,tol=1e-10):
        rc = self.FluxCorrCoef(reac1=reac,reac2=None,absolute=absolute,tol=tol)
        self.PrintSorted(rc, lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs, reverse=reverse)


    def StDev(self,reac):
        col = list(self[reac])
        mean = sum(col)/len(col)
        var = 0.0
        for b in col:
            var += (b-mean)**2
        var = var/len(col)
        return abs(math.sqrt(var))

    def PrintSD(self,lo=0,hi=float('inf'),f=None,Sort="value",sortabs=True,reverse=True):
        sd = {}
        for r in self.columns:
            sd[r] = self.StDev(r)
        self.PrintSorted(sd, lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs, reverse=reverse)

    def RelStDev(self,reac):
        col = list(self[reac])
        mean = sum(col)/len(col)
        var = 0.0
        for b in col:
            var += (b-mean)**2
        var = var/len(col)
        if mean == 0.0:
            return float('nan')
        else:
            return abs(math.sqrt(var)/mean)

    def PrintRSD(self,lo=0,hi=float('inf'),f=None,Sort="value",sortabs=True,reverse=True):
        rsd = {}
        for r in self.columns:
            rsd[r] = self.RelStDev(r)
        self.PrintSorted(rsd, lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs, reverse=reverse)

    def AverageDev(self,reac):
        col = list(self[reac])
        mean = sum(col)/len(col)
        var = 0.0
        for b in col:
            var += abs(b-mean)
        var = var/len(col)
        return var

    def PrintAD(self,lo=0,hi=float('inf'),f=None,Sort="value",sortabs=True,reverse=True):
        ad = {}
        for r in self.columns:
            ad[r] = self.AverageDev(r)
        self.PrintSorted(ad, lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs, reverse=reverse)

    def RelAverageDev(self,reac):
        col = list(self[reac])
        mean = sum(col)/len(col)
        var = 0.0
        for b in col:
            var += abs(b-mean)
        var = var/len(col)
        if mean == 0.0:
            return float('nan')
        else:
            var = var/abs(mean)
            return var

    def PrintRAD(self,lo=0,hi=float('inf'),f=None,Sort="value",sortabs=True,reverse=True):
        rad = {}
        for r in self.columns:
            rad[r] = self.RelAverageDev(r)
        self.PrintSorted(rad, lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs, reverse=reverse)

    def AsDic(self):
        rv = {}
        for r in self.index:
            row = self.loc[r,:]
            row = tuple(row)
            rv[r] = row
        return rv

    def PrintSorted(self, obj, lo=0, hi=float('inf'), f=None, Sort="value", sortabs=True, reverse=True):
        if f != None:
            temp = dict(obj)
            for reac in list(temp.keys()):
                if f not in reac:
                    del temp[reac]
            obj = temp
        if Sort == "value":
            if sortabs:
                function = lambda k,v: 0, k if math.isnan(v) else abs(v),k
            else:
                function = lambda k,v: 0, k if math.isnan(v) else v,k
            for key, value in sorted(obj.iteritems(), key=function, reverse=reverse):
                if hi >= abs(value) >= lo:
                    print("%s: %s" % (key, value))
        elif Sort == "key":
            for key in sorted(obj.iterkeys()):
                if hi >= abs(obj[key]) >= lo:
                    print("%s: %s" % (key, obj[key]))


from cobra import Reaction
#from cobra.manipulation import modify
from cobra.core.solution import Solution
from pandas import Series
from six import iteritems

def SplitRev(model, split_solution=True):
    #modify.convert_to_irreversible(model)
    reactions_to_add = []
    coefficients = {}
    for reaction in model.reactions:
        # If a reaction is reverse only, the forward reaction (which
        # will be constrained to 0) will be left in the model.
        if reaction.lower_bound < 0:
            reverse_reaction = Reaction(reaction.id + "_reverse")
            reverse_reaction.lower_bound = min(0, reaction.upper_bound) * -1
            reverse_reaction.upper_bound = reaction.lower_bound * -1
            coefficients[reverse_reaction] = reaction.objective_coefficient * -1
            reaction.lower_bound = 0
            reaction.upper_bound = max(0, reaction.upper_bound)
            #Make the directions aware of each other
            reaction.reflection = reverse_reaction
            reverse_reaction.reflection = reaction
            reaction.notes["reflection"] = reverse_reaction.id
            reverse_reaction.notes["reflection"] = reaction.id
            reaction_dict = {k: v * -1
                             for k, v in iteritems(reaction._metabolites)}
            reverse_reaction.add_metabolites(reaction_dict)
            reverse_reaction._model = reaction._model
            reverse_reaction._genes = reaction._genes
            for gene in reaction._genes:
                gene._reaction.add(reverse_reaction)
            reverse_reaction._gene_reaction_rule = reaction._gene_reaction_rule
            reverse_reaction.subsystem = reaction.subsystem
            reactions_to_add.append(reverse_reaction)
    model.add_reactions(reactions_to_add)
    model.SetObjective(coefficients)
    
    if split_solution:
        reverse_reactions = [x for x in model.reactions
                         if "reflection" in x.notes and
                         x.id.endswith('_reverse')]
        if isinstance(model.solution, Solution):
            fluxes = dict(model.solution.fluxes)
            reduced = dict(model.solution.reduced_costs)
            for reverse in reverse_reactions:
                forward = reverse.reflection
                if (forward.id in fluxes) and (reverse.id in fluxes):
                    if fluxes[forward.id] < 0:
                        fluxes[reverse.id] = -fluxes[forward.id]
                        fluxes[forward.id] = 0
                    else:
                        fluxes[reverse.id] = 0
                    if reduced[forward.id] < 0:
                        reduced[reverse.id] = -reduced[forward.id]
                        reduced[forward.id] = 0
                    else:
                        reduced[reverse.id] = 0
#                else:
#                    print(forward.id + ' and ' + reverse.id + ' not in solution')
            split_sol = Solution(model.solution.objective_value, model.solution.status,
                Series(index=fluxes.keys(), data=fluxes.values(), 
                name="fluxes"),
                Series(index=reduced.keys(), data=reduced.values(),
                name="reduced_costs"), model.solution.shadow_prices)
            model.UpdateSolution(split_sol)


def MergeRev(model, update_solution=True):
    ###modify.revert_to_reversible(model, update_solution=update_solution)

    """
        The cobra function above fails to update the solution object when merging. 
        This results in how reactions that goes in the reversible direction to be missing from solution. 
        Fix is made by commenting the remove_reactions function and, instead, changing the solution dict 
        manually in GetSol.
    """

    ### Copied cobra code here

    reverse_reactions = [x for x in model.reactions
                         if "reflection" in x.notes and
                         x.id.endswith('_reverse')]
    # If there are no reverse reactions, then there is nothing to do
    if len(reverse_reactions) == 0:
        return

    if update_solution:
        if isinstance(model.solution, Solution):
            fluxes = dict(model.solution.fluxes)
            reduced = dict(model.solution.reduced_costs)
            for reverse in reverse_reactions:
                forward = reverse.reflection
                if reverse.id in fluxes.keys():
                    fluxes[forward.id] -= fluxes.pop(reverse.id)
                    reduced[forward.id] -= reduced.pop(reverse.id)
            merged_sol = Solution(model.solution.objective_value, model.solution.status,
                            Series(index=fluxes.keys(), data=fluxes.values(), 
                            name="fluxes"),
                            Series(index=reduced.keys(), data=reduced.values(),
                            name="reduced_costs"), model.solution.shadow_prices)
            model.UpdateSolution(merged_sol)

    for reverse in reverse_reactions:
        forward_id = reverse.notes.pop("reflection")
        forward = model.reactions.get_by_id(forward_id)
        forward.lower_bound = -reverse.upper_bound
        if forward.upper_bound == 0:
            forward.upper_bound = -reverse.lower_bound

        if "reflection" in forward.notes:
            forward.notes.pop("reflection")
            
    model.remove_reactions(reverse_reactions)
    
    ###

    # Since the metabolites and genes are all still in
    # use we can do this faster removal step.  We can
    # probably speed things up here.
    #"""
    #This is commented out -> model.remove_reactions(reverse_reactions)
    #"""
    # reverse_reactions = [x for x in model.reactions
    #     if x.reflection is not None and x.id.endswith('_reverse')]
    # for reverse in reverse_reactions:
    #     forward = reverse.reflection
    #     forward.lower_bound = -reverse.upper_bound
    #     if reverse.lower_bound < 0:
    #         forward.upper_bound = -reverse.lower_bound
    #     forward.reflection = None
    # #Since the metabolites and genes are all still in
    # #use we can do this faster removal step.  We can
    # #probably speed things up here.
    # model.remove_reactions(reverse_reactions)
    # # fix the solution
    # if (update_solution) and (model.solution is not None) and (model.solution.x_dict is not None):
    #     x_dict = model.solution.x_dict
    #     for reverse in reverse_reactions:
    #         forward = reverse.reflection
    #         x_dict[forward.id] -= x_dict.pop(reverse.id)
    #     model.solution.x = [x_dict[r.id] for r in model.reactions]

#def MergeSolution(sol): 
#    """
#    This function takes in a Solution object that is unmerged (contains reactions both in the forward and #reverse direction)
#    and combines them to a single reaction.
#    """
#    old_fluxes = dict(sol_object.fluxes)
#    old_reduced = dict(sol_object.reduced_costs)
#    
#    new_fluxes = {}
#    new_reduced = {}
#
#    for reac in old_fluxes.keys(): 
#        if reac.endswith('_reverse'):
#            forward_id = reac[:-len('_reverse')] 
#            if old_fluxes[reac] == 0:
#                new_fluxes[forward_id] = old_fluxes[forward_id]
#            else: 
#                new_fluxes[forward_id] = - old_fluxes[reac]
#        else:
#            new_fluxes[reac] = old_fluxes[reac]
#    for reac in old_reduced.keys(): 
#        if reac.endswith('_reverse'): 
#            forward_id = reac[:-len('_reverse')]  
#            if old_reduced[reac] == 0:
#                new_reduced[forward_id] = old_reduced[forward_id]
#            else: 
#                new_reduced[forward_id] = - old_reduced[reac]
#        else:
#            new_reduced[reac] = old_reduced[reac]
#
#    return Solution(sol_object.objective_value, sol_object.status,
#                    Series(index=new_fluxes.keys(), data=new_fluxes.values(), name="fluxes"),
#                    Series(index=new_reduced.keys(), data=new_reduced.values(),
#                    name="reduced_costs"), sol_object.shadow_prices)



from .io.IO import ReadModel as Model
from .classes.model import model
from .classes.fva import fva
from .classes.fca import fca
from .classes.matrix import matrix
from .classes.pareto import pareto
from .classes.flux import flux
#from .classes.reaction import reaction
#from .classes.metabolite import metabolite
#from .classes.db import db
##import sys
##import os
##p = os.path.abspath(__file__)
##location = p[:p.rindex(os.sep)]
##sys.path.append(location)



import cobra
from fractions import Fraction

def ReadScrumPyModel(spy_file, compartment_dic={}, Print=False, AutoExt=True):
    """
    Take a ScrumPy model file generated in ScrumPy with m.smexterns.ToScrumPy()
    Cannot take a master model file with links to children files
    Does not work if the model file is not in the following format:

    Reaction name:
        5 metabolite_A + 10 metabolite_B -> 1 metabolite_C
        ~

    Note that each reaction must be separated into 3 lines
    """
    model_string = open(spy_file,'r').read()
    model_string = model_string.replace('\t','')
    model_string = model_string.replace('    ','')
    string_list = model_string.split('\n')

    model = cobra.Model()

    externals = []
    metabolite_dic = {}

    reac_name = None
    DeQuote = False
    for string in string_list:
        if string == '' or string.startswith("#"):
            pass
        elif string == 'Structural()':
            pass
        elif string == 'DeQuote()':
            DeQuote = True
        elif string.startswith('ElType(') and string.endswith(')'):
            pass
        elif string.startswith("External(") and string.endswith(")"):
            ext_string = string[9:-1]
            ext_string = ext_string.replace(' ','')
            ext_list = ext_string.split(',')
            externals.extend(ext_list)
        elif string.endswith(':'):
            reac_name = string[:-1]
        elif reac_name and string == '~':
            reac_name = None
        elif reac_name and string != '~':
            if DeQuote:
                if (reac_name.startswith('"') and reac_name.endswith('"')) or (reac_name.startswith("'") and reac_name.endswith("'")):
                    reac_name = reac_name[1:-1]
            reaction = cobra.Reaction(reac_name)
            stoichiometry_list = []
            metabolite_list = []
            new_metabolite = True
            product = False
            compartment = ''
            s_list = string.split(' ')
            for s in s_list:
                if s == ":" or s == "":
                    pass
                elif s == "+":
                    new_metabolite = True
                elif s == "->" or s == "-->" or s == "=>" or s == u'\u2192':
                    direction = 'LEFT-TO-RIGHT'
                    product = True
                    new_metabolite = True
                elif s == "<>" or s =="<=>" or s == "<==>" or s == "<->" or s == u'\u2194':
                    direction = 'REVERSIBLE'
                    product = True
                    new_metabolite = True
                elif s == "<-" or s == "<--" or s == "<=" or s == u'\u2190':
                    direction = 'RIGHT-TO-LEFT'
                    product = True
                    new_metabolite = True
                else:
                    try:
                        s = Fraction(s)
                        s = float(s)
                        if not product:
                            s = -s
                        stoichiometry_list.append(s)
                        new_metabolite = False
                    except ValueError:
                        metabolite_list.append(s + compartment)
                        if new_metabolite:
                            if not product:
                                stoichiometry_list.append(-1)
                            else:
                                stoichiometry_list.append(1)
                            new_metabolite = True
            stoi_dic = {}
            for b in range(len(metabolite_list)):
                met_id = metabolite_list[b]
                if met_id not in externals:
                    if (met_id.startswith('"') and met_id.endswith('"')) or (met_id.startswith("'") and met_id.endswith("'")):
                        dmet_id = met_id[1:-1]
                        if DeQuote:
                            met_id = met_id[1:-1]
                    else:
                        dmet_id = met_id
                    if AutoExt and (dmet_id.startswith('X_') or dmet_id.startswith('x_')):
                        externals.append(met_id)
                    else:
                        try:
                            met = metabolite_dic[met_id]
                        except KeyError:    # try case-insensitive
                            met_key = []
                            for key in metabolite_dic:
                                if key.lower() == met_id.lower():
                                    met_key.append(key)
                            if len(met_key) == 1:
                                met = metabolite_dic[met_key[0]]
                            else:
                                met_compartment = None
                                for c in compartment_dic:
                                    if met_id.endswith(c):
                                        met_compartment = compartment_dic[c]
                                met = cobra.Metabolite(met_id,compartment=met_compartment)
                                metabolite_dic[met_id] = met
                        stoi_dic[met] = stoichiometry_list[b]
            reaction.add_metabolites(stoi_dic)
            reaction.lower_bound = -float('inf')
            reaction.upper_bound = float('inf')
            if direction == 'LEFT-TO-RIGHT':
                reaction.lower_bound = 0.0
            elif direction == 'RIGHT-TO-LEFT':
                reaction.upper_bound = 0.0
            model.add_reaction(reaction)
            if Print:
                print(reaction.reaction)
    return model

def WriteScrumPyModel(model,filename, ExtReacs={}):
    """ ExtReacs = {reac_id:stoichiotry} """
    model = model.copy()
    arrow_dic = {'<--':'<-','<=>':'<>','-->':'->'}
    out_file = 'Structural()\n'
    if len(model.reactions) > 0:
        if not ((model.reactions[0].id.startswith('"') and model.reactions[0].id.endswith('"')) or (model.reactions[0].id.startswith("'") and model.reactions[0].id.endswith("'"))):
            out_file += 'DeQuote()\n'
    reactions_string = ''
    external_metabolites = []
    if isinstance(ExtReacs, str):
        ExtReacs = {ExtReacs:1}
    elif isinstance(ExtReacs, list):
        ExtReacs = dict((reac,1) for reac in ExtReacs)
    for met in model.metabolites:
        if not (met.id.startswith('"') and met.id.endswith('"')):
            met.id = '"' + met.id + '"'
    for reac in model.reactions:
        reactions_string += '"' + reac.id + '":\n'
        reac_string = reac.reaction
        for arrow in arrow_dic:
            reac_string = reac_string.replace(arrow,arrow_dic[arrow])
        if reac_string.startswith('<>') or reac_string.startswith('->') or reac_string.startswith(' <>') or reac_string.startswith(' ->') or reac_string.startswith('<-') or reac_string.startswith(' <-'):
            external_met = '"X_' + reac.id + '"'
            reactions_string += '    ' + external_met + reac_string + '\n    ~\n'
            external_metabolites.append(external_met)
        elif reac_string.endswith('<>') or reac_string.endswith('->') or reac_string.endswith('<> ') or reac_string.endswith('-> ') or reac_string.endswith('<-') or reac_string.endswith('<- '):
            external_met = '"X_' + reac.id + '"'
            reactions_string += '    ' + reac_string + external_met + '\n    ~\n'
            external_metabolites.append(external_met)
        elif reac.id in ExtReacs.keys():
            external_met = '"X_' + reac.id + '"'
            if ExtReacs[reac.id] > 0:
                reactions_string += '    ' + reac_string + ' + ' + str(abs(float(ExtReacs[reac.id]))) + ' ' + external_met + '\n    ~\n'
            elif ExtReacs[reac.id] < 0:
                reactions_string += '    ' + str(abs(float(ExtReacs[reac.id]))) + ' ' + external_met + ' + ' + reac_string + '\n    ~\n'
            external_metabolites.append(external_met)
        else:
            reactions_string += '    ' + reac_string + '\n    ~\n'
    if external_metabolites:
        ext_mets_string = ''
        for met in external_metabolites:
            ext_mets_string += str(met) + ','
        out_file += 'External(' + ext_mets_string[:-1] + ')\n'
    out_file += reactions_string
    open(filename,'w').write(out_file)





# -*- coding: utf-8 -*-
"""
Spyder Editor

This is a temporary script file.
"""
import cobra
import sys
import os

from ..classes.reaction import Reaction
from ..classes.metabolite import Metabolite

#FOR TESTING
#from reaction import Reaction
#from metabolite import Metabolite

def ReadCyc(reactionDatFile,compoundsDatFile="",classesDatFile="",enyzmeDatFile="",Print=False):
    """
    REQUIREMENTS TO USE: reactions.dat, compounds.dat, classes.dat--Either provide 
    the path or put them in the same directory and provide reactions.dat path.

    Read in reactions.dat file from cyc and return a cobra model.
    Returns
    -------
    scobra.Model
    """

    if(not reactionDatFile.endswith("reactions.dat")):
        raise Exception("Input error: not cyc reactions.dat file.")

    model = cobra.Model()
    
    #### KEYWORDS
    UID  = "UNIQUE-ID - "
    lUID= len(UID)
    
    # METABOLITE
    CN = "COMMON-NAME - "
    lCN = len(CN)

    CF = "CHEMICAL-FORMULA - "
    lCF = len(CF)

    AC = "ATOM-CHARGES - "
    lAC = len(AC)

    MW = "MOLECULAR-WEIGHT - "
    lMW = len(MW)

    INCHI = "INCHI - InChI="
    lINCHI = len(INCHI)

    SMILES = "SMILES - "
    lSMILES = len(SMILES)

    COMMENTS = "COMMENT - "
    lCOMMENTS = len(COMMENTS)

    # GENE
    ENZYME = "ENZYME - "
    lENZYME = len(ENZYME)

    REACTION = "REACTION -"
    lREACTION = len(REACTION)

    # REACTION
    L = "LEFT - "
    lL= len(L)
    R = "RIGHT - "
    lR= len(R)
    CO = "^COEFFICIENT - "
    lCO = len(CO)
    
    RD = "REACTION-DIRECTION - "
    lRD = len(RD)
    ltr  = "LEFT-TO-RIGHT"
    rev = "REVERSIBLE"
    rtl = "RIGHT-TO-LEFT"

    ENZ_REC = "ENZYMATIC-REACTION - "
    lENZ_REC = len(ENZ_REC)
    
    #### READING IN METABOLITES FROM COMPOUNDS ####
    metStreamCompounds = None
    
    try:
        if compoundsDatFile == "":
            metStreamCompounds = open(reactionDatFile.replace("reactions.dat","compounds.dat"), 'r', encoding="utf8", errors='ignore')
        else:
            metStreamCompounds = open(compoundsDatFile, 'r', encoding="utf8", errors='ignore')
    except:
        print("compounds.dat is required for importing metabolites but is not provided/found in the directory.")
        return
    #TESTFORMULA = []
    mets_dict = {}
    line = metStreamCompounds.readline()
    metabolite = Metabolite("")
    while(line):
        line = line.rstrip("\n")
        if(line.startswith(UID)):
            if(metabolite.id!=""):
                mets_dict[metabolite.id] = metabolite
            metabolite = Metabolite(id=line[lUID:].strip(), name=line[lUID:].strip())
            metabolite.charge = 0
        elif(line.startswith(CN)):
            metabolite.name = line[lCN:].strip()

        elif(line.startswith(AC)):
            charge=0
            while(line.startswith(AC)):
                chargeStr = line[lAC:].strip().strip("()").split(" ")[1]
                charge+=int(chargeStr)
                line = metStreamCompounds.readline()
            metabolite.charge=charge
            continue

        elif(line.startswith(CF)):
            formula = ""
            while(line.startswith(CF)):
                chemString = line[lCF:].strip().strip("()").split(" ")
                chemF = chemString[0]+chemString[1]
                formula = formula + chemF
                line = metStreamCompounds.readline()
            #TESTFORMULA.append(formula)
            metabolite.formula=formula
            continue

        elif(line.startswith(MW)):
            metabolite.molecular_weights = float(line[lMW:].strip())

        elif(line.startswith(SMILES)):
            metabolite.smiles = line[lSMILES:].strip()

        line = metStreamCompounds.readline()

    metStreamCompounds.close()


    #### READING IN METABOLITES FROM CLASSES ####
    metStreamClasses = None
    try:
        if classesDatFile == "":
            #print("first one")
            metStreamClasses = open(reactionDatFile.replace("reactions.dat","classes.dat"), 'r', encoding="utf8", errors='ignore')
        else:
            metStreamClasses = open(classesDatFile, 'r', encoding="utf8", errors='ignore')
    except:
        print("classes.dat is required for importing metabolites but is not provided/found in the directory.")

    line = metStreamClasses.readline()
    metabolite = Metabolite("")
    while(line):
        line = line.rstrip("\n")
        if(line.startswith(UID)):
            if(metabolite.id!=""):
                mets_dict[metabolite.id] = metabolite
            metabolite = Metabolite(id=line[lUID:].strip())
            metabolite.charge = 0
        elif(line.startswith(CN)):
            metabolite.name = line[lCN:].strip()

        elif(line.startswith(COMMENTS)):
            metabolite.COMMENTS = line[lCOMMENTS:].strip()

        line = metStreamClasses.readline()

    metStreamClasses.close()

    #### READING IN ENZYMES ####
    enzymeStream = None
    try:
        if enyzmeDatFile == "":
            #print("first one")
            enzymeStream = open(reactionDatFile.replace("reactions.dat","enzrxns.dat"), 'r', encoding="utf8", errors='ignore')
        else:
            enzymeStream = open(enyzmeDatFile, 'r', encoding="utf8", errors='ignore')
    except:
        print("enzrxns.dat is required for importing metabolites but is not provided/found in the directory.")

    line = enzymeStream.readline()

    enzs_dict = {}

    enz_id = ""
    enz_name = ""
    while(line):
        line = line.rstrip("\n")
        if(line.startswith(UID)):
            if(enz_id!=""):
                enzs_dict[enz_id] = enz_name
            enz_id = line[lUID:].strip()
        elif(line.startswith(ENZYME)):
            enz_name = line[lENZYME:].strip()

        line = enzymeStream.readline()

    enzymeStream.close()

    #### READING IN REACTION ####
    reactionStream = open(reactionDatFile, 'r', encoding="utf8", errors='ignore')

    line = reactionStream.readline()
    #all_stoic_ = []
    #all_reacs_ = []
    #all_dirs_ = []
    
    #model.metabolites = mets_dict

    reaction = Reaction("")
    stoi_dict = {}
    added = 0
    proteins = {}

    while(line):
        line=line.rstrip("\n")
        if(line.startswith(UID)):
            if(reaction.id!=""):
                #print(stoi_dict)
                reaction.add_metabolites(stoi_dict,reversibly=(direction==rev))
                #added+=1
                #print(added)
                #all_stoic_.append(stoi_dict)
                model.add_reaction(reaction)
            stoi_dict = {}
            reaction = Reaction(line[lUID:].strip())
            #all_reacs_.append(line[lUID:].strip())
            
        elif(line.startswith(ENZ_REC)):
            enz_id = line[lENZ_REC:]
            reaction.proteins[enz_id] = enzs_dict[enz_id]

        elif(line.startswith(L)):
            met = line[lL:]
            if met in mets_dict:
                met = mets_dict[met]
            else:
                met = Metabolite(id = met, name = met)
            stoi_dict[met] = -1
            line = reactionStream.readline()
            if(line.startswith(CO)):
                try:
                    int(line[lCO:])
                except ValueError:
                    line = reactionStream.readline()
                    continue
                stoi_dict[met] = int(line[lCO:])
                line = reactionStream.readline()
            continue

        elif(line.startswith(RD)):
            d = line[lRD:]
            direction = ""
            if(ltr in d):
                direction = ltr
            elif(rtl in d):
                direction = rtl
            elif(rev in d):
                direction = rev
            else:
                raise Exception("invalid direction for: "+reaction.id)
            #all_dirs_.append(direction)

        elif(line.startswith(R)):
            met = line[lR:]
            if met in mets_dict:
                met = mets_dict[met]
            else:
                met = Metabolite(id = met, name = met)
            stoi_dict[met] = 1
            line = reactionStream.readline()
            if(line.startswith(CO)):
                try:
                    int(line[lCO:])
                except ValueError:
                    line = reactionStream.readline()
                    continue
                stoi_dict[met] = int(line[lCO:])
                line = reactionStream.readline()
            continue
        
        line = reactionStream.readline()
    #print(all_stoic_[2])
    #print(all_dirs_[2])
    #print(all_reacs_[2])
    #print(TESTFORMULA)
    #Accounting for last reaction here:
    reaction.add_metabolites(stoi_dict,reversibly=(direction==rev))
    #added+=1
    #all_stoic_.append(stoi_dict)
    model.add_reaction(reaction)

    #print(model.metabolites)
    #print(model.reactions)

    reactionStream.close()

    return model
